from pathlib import Path


BASE_DIR = Path(__file__).resolve().parent.parent
ALLOWED_HOSTS = ["localhost", "127.0.0.1"]
DEBUG = True
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"
LANGUAGE_CODE = "en-us"
MEDIA_ROOT = BASE_DIR / "media"
MEDIA_URL = "media/"
DOCS_ROOT = BASE_DIR / "docs" / "build" / "html"
ROOT_URLCONF = "src.urls"
SECRET_KEY = "django-insecure-#ezlo7tqc&h07y4g^1i3jqg78^z*jgsyd11kq812^=k4%!lk6b"
STATIC_URL = "static/"
STATICFILES_DIRS = [BASE_DIR / "static"]
TIME_ZONE = "America/Chicago"
USE_I18N = True
USE_TZ = True
WSGI_APPLICATION = "src.wsgi.application"
SILENCED_SYSTEM_CHECKS = ["staticfiles.W004"]
S3_UPLOAD_BUCKET_NAME = "crimsonslate-bucket"
S3_UPLOAD_REGION = "us-east-1"


PORTFOLIO_PROFILE = {
    "NAME": "Crimsonslate",
    "FIRST_NAME": "Crimson",
    "LAST_NAME": "Slate",
    "EMAIL": "contact@crimsonslate.com",
    "PHONE": "+15555555555",
    "SOCIALS": {
        "DISCORD": {
            "display_name": "crimsonslate",
            "profile_link": "https://discord.gg/***",  # Server link
            "username": "crimsonslate",
        },
        "YOUTUBE": {
            "display_name": "crimsonslate",
            "profile_link": "https://youtube.com/***",
            "username": "crimsonslate",
        },
        "INSTAGRAM": {
            "display_name": "crimsonslate",
            "profile_link": "https://instagram.com/@***",
            "username": "crimsonslate",
        },
        "FACEBOOK": {
            "display_name": "crimsonslate",
            "profile_link": "https://facebook.com/***",
            "username": "crimsonslate",
        },
        "TIKTOK": {
            "display_name": "crimsonslate",
            "profile_link": "https://facebook.com/***",
            "username": "crimsonslate",
        },
        "TWITTER": {
            "display_name": "crimsonslate",
            "profile_link": "https://x.com/@***",
            "username": "crimsonslate",
        },
        "REDDIT": {
            "display_name": "crimsonslate",
            "profile_link": "https://reddit.com/u/***",
            "username": "crimsonslate",
        },
    },
}

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    },
}

STORAGES = {
    "default": {
        "BACKEND": "django.core.files.storage.FileSystemStorage",
    },
    "staticfiles": {
        "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
    },
    "bucket": {
        "BACKEND": "storages.backends.s3.S3Storage",
        "OPTIONS": {
            "location": "media/",
            "bucket_name": S3_UPLOAD_BUCKET_NAME,
            "region_name": S3_UPLOAD_REGION,
            "verify": False,
        },
    },
}

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.admindocs",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.forms",
    "docs",
    "crimsonslate_portfolio.apps.CrimsonslatePortfolioConfig",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]


TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [BASE_DIR / "templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

FILE_UPLOAD_HANDLERS = [
    "django.core.files.uploadhandler.MemoryFileUploadHandler",
    "django.core.files.uploadhandler.TemporaryFileUploadHandler",
]

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]



from typing import Any

from django.template import Library

from crimsonslate_portfolio.models import Media

register = Library()


@register.inclusion_tag("portfolio/media/display.html")
def display(
    media: Media, force_image: bool = False, css_class: str | None = None
) -> dict[str, Any]:
    return {
        "title": media.title,
        "class": css_class,
        "image": force_image if force_image else media.is_image,
        "src": media.source.url,
        "detail_url": media.get_absolute_url(),
    }



from django.contrib.auth.forms import AuthenticationForm
from django.forms import ModelForm

from crimsonslate_portfolio.models import Media


class PortfolioAuthenticationForm(AuthenticationForm):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        for name in self.fields:
            self.fields[name].widget.attrs.update(
                {"class": "p-2 rounded bg-white", "placeholder": name.title()}
            )


class MediaCreateForm(ModelForm):
    class Meta:
        model = Media
        fields = [
            "source",
            "thumb",
            "title",
            "subtitle",
            "desc",
            "is_hidden",
            "tags",
            "date_created",
        ]


class MediaUpdateForm(ModelForm):
    class Meta:
        model = Media
        fields = [
            "source",
            "thumb",
            "title",
            "subtitle",
            "desc",
            "is_hidden",
            "tags",
            "date_created",
        ]



from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.urls import path, include

urlpatterns = [
    path("admin/", admin.site.urls),
    path("admin/docs/", include("django.contrib.admindocs.urls")),
    path("docs/", include("docs.urls")),
    path("", include("crimsonslate_portfolio.urls")),
]

if settings.DEBUG:
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)
    urlpatterns += static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)



from django.core.files import File
from django.core.validators import (
    FileExtensionValidator,
    get_available_image_extensions,
)


def validate_media_file_extension(value: File) -> None:
    video_extensions: list[str] = ["mp4"]  # Only mp4 is supported rn
    image_extensions: list[str] = list(get_available_image_extensions())
    validator = FileExtensionValidator(
        allowed_extensions=[
            file_extension for file_extension in video_extensions + image_extensions
        ]
    )
    validator(value)
    return



from django.apps import AppConfig


class CrimsonslatePortfolioConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "crimsonslate_portfolio"
    verbose_name = "Crimsonslate Portfolio Django Application"



from django.contrib.auth.mixins import LoginRequiredMixin
from django.urls import reverse_lazy
from django.views.generic import (
    CreateView,
    DeleteView,
    DetailView,
    TemplateView,
    UpdateView,
    ListView,
)

from crimsonslate_portfolio.forms import MediaCreateForm, MediaUpdateForm
from crimsonslate_portfolio.models import Media
from crimsonslate_portfolio.views.mixins import (
    HtmxTemplateResponseMixin,
    PortfolioProfileMixin,
    PortfolioSingleObjectMixin,
    PortfolioMultipleObjectMixin,
)


class MediaUploadView(
    LoginRequiredMixin, PortfolioProfileMixin, HtmxTemplateResponseMixin, TemplateView
):
    http_method_names = ["get"]
    extra_context = {"title": "Upload", "class": "bg-violet-300"}
    template_name = "portfolio/media/upload.html"
    partial_template_name = "portfolio/media/partials/_upload.html"
    login_url = reverse_lazy("login")
    raise_exception = False
    permission_denied_message = "Please login and try again."


class MediaUploadSuccessView(
    HtmxTemplateResponseMixin, PortfolioProfileMixin, TemplateView
):
    http_method_names = ["get"]
    extra_context = {"title": "Success", "class": ""}
    template_name = "portfolio/media/upload_success.html"
    partial_template_name = "portfolio/media/partials/_upload_success.html"


class MediaDeleteView(
    HtmxTemplateResponseMixin,
    PortfolioSingleObjectMixin,
    PortfolioProfileMixin,
    DeleteView,
):
    model = Media
    template_name = "portfolio/media/delete.html"
    partial_template_name = "portfolio/media/partials/_delete.html"
    http_method_names = ["get", "post"]


class MediaCreateView(HtmxTemplateResponseMixin, PortfolioProfileMixin, CreateView):
    form_class = MediaCreateForm
    http_method_names = ["get", "post"]
    model = Media
    partial_template_name = "portfolio/media/partials/_create.html"
    template_name = "portfolio/media/create.html"


class MediaUpdateView(
    HtmxTemplateResponseMixin,
    PortfolioSingleObjectMixin,
    PortfolioProfileMixin,
    UpdateView,
):
    form_class = MediaUpdateForm
    http_method_names = ["get", "post"]
    model = Media
    partial_template_name = "portfolio/media/partials/_update.html"
    template_name = "portfolio/media/update.html"


class MediaDetailView(
    HtmxTemplateResponseMixin,
    PortfolioSingleObjectMixin,
    PortfolioProfileMixin,
    DetailView,
):
    http_method_names = ["get"]
    model = Media
    partial_template_name = "portfolio/media/partials/_detail.html"
    queryset = Media.objects.all()
    template_name = "portfolio/media/detail.html"


class MediaGalleryView(
    HtmxTemplateResponseMixin,
    PortfolioMultipleObjectMixin,
    PortfolioProfileMixin,
    ListView,
):
    allow_empty = True
    extra_context = {"title": "Gallery"}
    http_method_names = ["get"]
    model = Media
    ordering = "date_created"
    paginate_by = 12
    partial_template_name = "portfolio/media/partials/_gallery.html"
    queryset = Media.objects.all()
    template_name = "portfolio/media/gallery.html"


class MediaSearchView(
    HtmxTemplateResponseMixin,
    PortfolioMultipleObjectMixin,
    PortfolioProfileMixin,
    ListView,
):
    allow_empty = True
    extra_context = {"title": "Search"}
    http_method_names = ["get"]
    model = Media
    ordering = "name"
    paginate_by = 12
    partial_template_name = "portfolio/media/partials/_search.html"
    queryset = Media.objects.all()
    template_name = "portfolio/media/search.html"



from django.conf import settings
from django.conf.urls.static import static
from django.urls import path

from . import views

urlpatterns = [
    path("login/", views.LoginView.as_view(), name="login"),
    path("logout/", views.LogoutView.as_view(), name="logout"),
    path("contact/", views.ContactView.as_view(), name="contact"),
    path("gallery/", views.MediaGalleryView.as_view(), name="gallery"),
    path("search/", views.MediaSearchView.as_view(), name="search"),
    path("upload/", views.MediaUploadView.as_view(), name="upload media"),
    path(
        "upload/success/", views.MediaUploadSuccessView.as_view(), name="upload success"
    ),
    path("media-creation/", views.MediaCreateView.as_view(), name="create media"),
    path("<str:slug>/", views.MediaDetailView.as_view(), name="detail media"),
    path("<str:slug>/delete/", views.MediaDeleteView.as_view(), name="delete media"),
    path("<str:slug>/update/", views.MediaUpdateView.as_view(), name="update media"),
]

if settings.DEBUG:
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)
    urlpatterns += static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)



import argparse
import json
import os

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    help = "Runs the tailwind compiler"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        """
        Adds subcommand arguments to the ``tailwind`` command.

        +-------------+------------------------------------------------------------------+
        | Subcommand  | Action                                                           |
        +=============+==================================================================+
        | ``install`` | Installs the tailwind compiler for the project.                  |
        +-------------+------------------------------------------------------------------+
        | ``start``   | Starts the tailwind compiler. Can be canceled with ``<CTRL>-c``. |
        +-------------+------------------------------------------------------------------+
        | ``build``   | Builds the tailwind output file for production.                  |
        +-------------+------------------------------------------------------------------+

        :param parser: An argument parser.
        :type parser: :py:obj:`argparse.ArgumentParser`
        :returns: Nothing.
        :rtype: :py:obj:`None`

        """
        subparsers = parser.add_subparsers(dest="subcommand")
        subparsers.add_parser("install", help="Install tailwind compiler")
        subparsers.add_parser("start", help="Start tailwind compiler")
        subparsers.add_parser("build", help="Compile tailwind for production")

    def handle(self, *args, **options):
        """
        Handles command execution based on the provided subcommand.

        :raises CommandError: If the subcommand is invalid.
        :returns: Nothing.
        :rtype: :py:obj:`None`

        """
        subcommand = options["subcommand"]
        try:
            command = self.generate_command(subcommand)
            os.system(command)
        except ValueError as e:
            raise CommandError(e)

    def generate_command(self, subcommand: str | None) -> str:
        """
        Generates a tailwind command based on the provided subcommand.

        :returns: A command to be executed by :py:func:`os.system`.
        :rtype: :py:obj:`str`

        """
        input, output = self.get_input_filepath(), self.get_output_filepath()
        command = f"npx @tailwindcss/cli -i {input} -o {output} "

        match subcommand:
            case "start":
                styled = self.style.NOTICE
                message = "Starting tailwind compiler..."
                command += "--watch"
            case "build":
                styled = self.style.NOTICE
                message = "Compiling tailwind for production..."
                command += "--minify"
            case "install":
                if self.node_package_installed("tailwind"):
                    styled = self.style.WARNING
                    message = "Tailwind is already installed, building for production instead..."
                    command += "--minify"
                else:
                    styled = self.style.NOTICE
                    message = "Installing tailwind..."
                    command = "npm install -D tailwindcss @tailwindcss/cli"
            case _:
                raise ValueError(f"Invalid subcommand: '{subcommand}'.")

        self.stdout.write(styled(message))
        return command

    def get_node_dependencies(self) -> list[str]:
        """
        Retrives a list of npm dependencies from ``package.json``.

        Returns an empty list if ``package.json`` does not exist.

        :returns: A list of application dependencies as strings.
        :rtype: :py:obj:`list`

        """
        dependencies: list[str] = []
        if not os.path.isfile("package.json"):
            return dependencies

        with open("package.json", "r") as file:
            dependencies.extend(json.load(file).get("devDependencies", {}).keys())
        return dependencies

    def node_package_installed(self, name: str) -> bool:
        """
        Whether or not a node package is installed.

        :type: :py:obj:`bool`

        """
        return name in self.get_node_dependencies()

    def get_input_filepath(self) -> str:
        """Returns the first input.css file found in ``BASE_DIR/css``"""
        return list(settings.BASE_DIR.glob("**/css/input.css"))[0]

    def get_output_filepath(self) -> str:
        """Returns the first output.css file found in ``BASE_DIR/css``"""
        return list(settings.BASE_DIR.glob("**/css/output.css"))[0]



from datetime import date

from django.core.files.storage import storages
from django.core.validators import (
    get_available_image_extensions,
    validate_image_file_extension,
)
from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.text import slugify

from crimsonslate_portfolio.validators import validate_media_file_extension


class MediaTag(models.Model):
    name = models.CharField(max_length=64)
    emoji = models.CharField(max_length=1, null=True, blank=True, default=None)

    class Meta:
        ordering = ["name"]
        verbose_name = "tag"
        verbose_name_plural = "tags"

    def __str__(self) -> str:
        return str(self.name)


class Media(models.Model):
    title = models.CharField(
        max_length=64,
        unique=True,
    )
    source = models.FileField(
        storage=storages["bucket"],
        upload_to="source/",
        validators=[validate_media_file_extension],
    )
    thumb = models.ImageField(
        storage=storages["bucket"],
        upload_to="thumb/",
        validators=[validate_image_file_extension],
        verbose_name="thumbnail",
    )
    subtitle = models.CharField(max_length=128, blank=True, null=True, default=None)
    desc = models.TextField(
        verbose_name="description", max_length=2048, blank=True, null=True, default=None
    )
    slug = models.SlugField(
        max_length=64, unique=True, blank=True, null=True, default=None
    )
    is_hidden = models.BooleanField(default=False)
    is_image = models.BooleanField(blank=True, null=True, editable=False)
    tags = models.ManyToManyField("MediaTag", default=None, blank=True)

    date_created = models.DateField(default=date.today)
    datetime_published = models.DateTimeField(default=timezone.now, editable=False)

    class Meta:
        ordering = ["-date_created"]
        constraints = [
            models.UniqueConstraint(
                fields=["title", "slug"],
                name="%(app_label)s_%(class)s_unique_title_and_slug",
            )
        ]

    def __str__(self) -> str:
        return str(self.title)

    def save(self, **kwargs) -> None:
        self.is_image = self.file_extension in get_available_image_extensions()
        if not self.slug or self.slug != slugify(self.title):
            self.slug = slugify(self.title)
        return super().save(**kwargs)

    def get_absolute_url(self) -> str:
        return reverse("detail media", kwargs={"slug": self.slug})

    @property
    def file_extension(self) -> str:
        return self.source.file.name.split(".")[-1]



from django.contrib.auth.views import LoginView as LoginViewBase
from django.contrib.auth.views import LogoutView as LogoutViewBase
from django.urls import reverse_lazy
from django.views.generic import TemplateView

from crimsonslate_portfolio.forms import PortfolioAuthenticationForm
from crimsonslate_portfolio.views.mixins import (
    HtmxTemplateResponseMixin,
    PortfolioProfileMixin,
)


class ContactView(PortfolioProfileMixin, HtmxTemplateResponseMixin, TemplateView):
    content_type = "text/html"
    extra_context = {"title": "Contact"}
    http_method_names = ["get", "post"]
    partial_template_name = "portfolio/partials/_contact.html"
    template_name = "portfolio/contact.html"


class LoginView(HtmxTemplateResponseMixin, LoginViewBase):
    content_type = "text/html"
    extra_context = {"title": "Login"}
    form_class = PortfolioAuthenticationForm
    http_method_names = ["get", "post"]
    partial_template_name = "portfolio/partials/_login.html"
    redirect_authenticated_user = True
    success_url = reverse_lazy("gallery")
    template_name = "portfolio/login.html"


class LogoutView(HtmxTemplateResponseMixin, LogoutViewBase):
    content_type = "text/html"
    extra_context = {"title": "Logout"}
    http_method_names = ["get", "post"]
    partial_template_name = "portfolio/partials/_logout.html"
    success_url = reverse_lazy("gallery")
    template_name = "portfolio/logout.html"



from .generic import ContactView, LoginView, LogoutView
from .media import (
    MediaCreateView,
    MediaDeleteView,
    MediaDetailView,
    MediaGalleryView,
    MediaSearchView,
    MediaUpdateView,
    MediaUploadView,
    MediaUploadSuccessView,
)



from typing import Any

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models import QuerySet
from django.views.generic.base import ContextMixin, TemplateResponseMixin
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.list import MultipleObjectMixin

if not hasattr(settings, "PORTFOLIO_PROFILE"):
    raise ImproperlyConfigured("'PORTFOLIO_PROFILE' setting is required.")


class HtmxTemplateResponseMixin(TemplateResponseMixin):
    """A template mixin that enables htmx features."""

    partial_template_name: str | None = None
    """
    A partial template rendered by htmx.

    :type: :py:obj:`str` | :py:obj:`None`
    :value: :py:obj:`None`

    """

    def render_to_response(self, context: dict[str, Any], **response_kwargs):
        """
        Sets :py:attr:`template_name` to :py:attr:`partial_template_name` if it is present.

        The request must be an HTMX request and not `boosted`_.

        .. _boosted: https://htmx.org/attributes/hx-boost/
        """
        htmx_request = bool(self.request.headers.get("HX-Request"))
        boosted = bool(self.request.headers.get("HX-Boosted"))

        if htmx_request and self.partial_template_name and not boosted:
            self.template_name = self.partial_template_name
        return super().render_to_response(context, **response_kwargs)


class PortfolioProfileMixin(ContextMixin):
    """Adds :confval:`PORTFOLIO_PROFILE` to the view context."""

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        context: dict[str, Any] = super().get_context_data(**kwargs)
        context["profile"] = settings.PORTFOLIO_PROFILE
        return context


class PortfolioSingleObjectMixin(SingleObjectMixin):
    def get_queryset(self) -> QuerySet:
        queryset = super().get_queryset()

        user = self.request.user if hasattr(self, "request") else None
        if not user or not user.is_staff:
            return queryset.exclude(is_hidden=True)
        return queryset


class PortfolioMultipleObjectMixin(MultipleObjectMixin):
    def get_queryset(self) -> QuerySet:
        queryset = super().get_queryset()

        user = self.request.user if hasattr(self, "request") else None
        if not user or not user.is_staff:
            return queryset.exclude(is_hidden=True)
        return queryset
